import xml.etree.ElementTree as ET
from typing import List, Tuple

import numpy as np
import pytest
from bioio_base import dimensions, exceptions, test_utilities

from bioio_czi import Reader

from .conftest import LOCAL_RESOURCES_DIR


@pytest.mark.parametrize(
    "filename, "
    "set_scene, "
    "expected_scenes, "
    "expected_shape, "
    "expected_dtype, "
    "expected_dims_order, "
    "expected_channel_names, "
    "expected_physical_pixel_sizes",
    [
        (
            "S=2_4x2_T=2=Z=3_CH=2.czi",
            "TR1",
            ("TR1", "TR2"),
            (2, 2, 3, 487, 947),
            np.uint16,
            "TCZYX",
            ["DAPI", "EGFP"],
            (1.0, 0.4, 0.4),
        ),
        (
            "s_1_t_1_c_1_z_1.czi",
            "Image:0",
            ("Image:0",),
            (1, 325, 475),
            np.uint16,
            "CYX",
            ["Bright"],
            (None, 1.0833333333333333, 1.0833333333333333),
        ),
        (
            "s_3_t_1_c_3_z_5.czi",
            "P2",
            ("P2", "P3", "P1"),
            (3, 5, 325, 475),
            np.uint16,
            "CZYX",
            [
                "EGFP",
                "TaRFP",
                "Bright",
            ],
            (1.0, 1.0833333333333333, 1.0833333333333333),
        ),
        (
            "s_3_t_1_c_3_z_5.czi",
            "P3",
            ("P2", "P3", "P1"),
            (3, 5, 325, 475),
            np.uint16,
            "CZYX",
            [
                "EGFP",
                "TaRFP",
                "Bright",
            ],
            (1.0, 1.0833333333333333, 1.0833333333333333),
        ),
        (
            "s_3_t_1_c_3_z_5.czi",
            "P1",
            ("P2", "P3", "P1"),
            (3, 5, 325, 475),
            np.uint16,
            "CZYX",
            [
                "EGFP",
                "TaRFP",
                "Bright",
            ],
            (1.0, 1.0833333333333333, 1.0833333333333333),
        ),
        (
            "RGB-8bit.czi",
            "Image:0",
            ("Image:0",),
            (624, 924, 3),  # bioio-czi has (1, 624, 924, 3),
            np.uint8,
            "YXS",  # bioio-czi has "TYXS",
            None,
            (None, 1.0833333333333333, 1.0833333333333333),
        ),
        (
            "variable_per_scene_dims.czi",
            "P2-D4",
            ("P1-D4", "P2-D4"),
            # The dimensions of the second scene are actually (1, 1, 2, 1248, 1848),
            # but pylibczirw doesn't make that metadata available and instead assumes
            # all scenes have the same shape.
            # This is a known defect of pylibczirw mode.
            (2, 1, 2, 1248, 1848),
            np.uint16,
            "TCZYX",
            ["CMDRP"],
            (2.23, 0.5416666666666666, 0.5416666666666666),
        ),
        pytest.param(
            "variable_scene_shape_first_scene_pyramid.czi",
            "A1",
            ("A1", "A2"),
            (3, 9, 2208, 2752),
            np.uint16,
            "CMYX",
            [
                "EGFP",
                "mCher",
                "PGC",
            ],
            (None, 9.082107048835329, 9.082107048835329),
            marks=pytest.mark.xfail(reason="Missing scenes"),
        ),
        pytest.param(
            "example.txt",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.UnsupportedFileFormatError),
        ),
        pytest.param(
            "s_1_t_1_c_1_z_1.ome.tiff",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=exceptions.UnsupportedFileFormatError),
        ),
    ],
)
def test_czi_reader(
    filename: str,
    set_scene: str,
    expected_scenes: Tuple[str, ...],
    expected_shape: Tuple[int, ...],
    expected_dtype: np.dtype,
    expected_dims_order: str,
    expected_channel_names: List[str],
    expected_physical_pixel_sizes: Tuple[float, float, float],
) -> None:
    # Construct full filepath
    uri = LOCAL_RESOURCES_DIR / filename

    # Run checks
    test_utilities.run_image_file_checks(
        ImageContainer=Reader,
        image=uri,
        set_scene=set_scene,
        expected_scenes=expected_scenes,
        expected_current_scene=set_scene,
        expected_shape=expected_shape,
        expected_dtype=expected_dtype,
        expected_dims_order=expected_dims_order,
        expected_channel_names=expected_channel_names,
        expected_physical_pixel_sizes=expected_physical_pixel_sizes,
        expected_metadata_type=ET.Element,
    )


@pytest.mark.parametrize(
    "url, expected_shape",
    [
        (
            "https://allencell.s3.amazonaws.com/aics/hipsc_12x_overview_image_dataset/"
            "stitchedwelloverviewimagepath/05080558_3500003720_10X_20191220_D3.czi"
            "?versionId=_KYMRhRvKxnu727ssMD2_fZD5CmQMNw6",
            (1, 5684, 5925),
        ),
    ],
)
def test_czi_reader_remote(url: str, expected_shape: Tuple[int]) -> None:
    # Construct full filepath
    reader = Reader(url)
    assert reader.shape == expected_shape


@pytest.mark.parametrize(
    "filename, expected_y_coords, expected_x_coords",
    [
        (
            "OverViewScan.czi",
            np.arange(0, 14692.852709554172, 4.5743626119409),
            np.arange(0, 33836.560240526844, 4.5743626119409),
        ),
    ],
)
def test_czi_reader_mosaic_coords(
    filename: str, expected_y_coords: np.ndarray, expected_x_coords: np.ndarray
) -> None:
    # Construct full filepath
    uri = LOCAL_RESOURCES_DIR / filename

    # Construct reader
    reader = Reader(uri)

    # Different from bioio-czi reader: the size of xarray_dask_data is the size of
    # a bounding box around all tiles in the scene, not just a single tile. (at the
    # lowest pyramid level (highest resolution)).
    # (at the lowest pyramid level (highest resolution)
    np.testing.assert_array_equal(
        reader.xarray_dask_data.coords[dimensions.DimensionNames.SpatialY].data,
        expected_y_coords,
    )
    np.testing.assert_array_equal(
        reader.xarray_dask_data.coords[dimensions.DimensionNames.SpatialX].data,
        expected_x_coords,
    )

    # Same as bioio-czi reader: mosaic dimensions are the full size of the lowest
    # pyramid level (highest resolution).
    np.testing.assert_array_equal(
        reader.mosaic_xarray_dask_data.coords[dimensions.DimensionNames.SpatialY].data,
        expected_y_coords,
    )
    np.testing.assert_array_equal(
        reader.mosaic_xarray_dask_data.coords[dimensions.DimensionNames.SpatialX].data,
        expected_x_coords,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import sys
import traceback
from pathlib import Path

from quilt3 import Package

from bioio_czi import __version__

###############################################################################

logging.basicConfig(
    level=logging.INFO,
    format="[%(levelname)4s: %(module)s:%(lineno)4s %(asctime)s] %(message)s",
)
log = logging.getLogger(__name__)

###############################################################################
# Args


class Args(argparse.Namespace):
    def __init__(self):
        self.__parse()

    def __parse(self):
        # Setup parser
        p = argparse.ArgumentParser(
            prog="upload_test_resources",
            description=(
                "Upload files used for testing this project. This will upload "
                "whatever files are currently found in the `tests/resources` directory."
                "To add more test files, simply add them to the `tests/resources` "
                "directory and rerun this script."
            ),
        )

        # Arguments
        p.add_argument(
            "--dry-run",
            action="store_true",
            help=(
                "Conduct dry run of the package generation. Will create a JSON "
                "manifest file of that package instead of uploading."
            ),
        )
        p.add_argument(
            "-y",
            "--yes",
            action="store_true",
            dest="preapproved",
            help="Auto-accept upload of files.",
        )
        p.add_argument(
            "--debug",
            action="store_true",
            help="Show traceback if the script were to fail.",
        )

        # Parse
        p.parse_args(namespace=self)


###############################################################################
# Build package


def upload_test_resources(args: Args):
    # Try running the download pipeline
    try:
        # Get test resources dir
        resources_dir = (
            Path(__file__).parent.parent / "bioio_czi" / "tests" / "resources"
        ).resolve(strict=True)

        # Report with directory will be used for upload
        log.info(f"Using contents of directory: {resources_dir}")

        # Create quilt package
        package = Package()
        package.set_dir("resources", resources_dir)

        # Report package contents
        log.info(f"Package contents: {package}")

        # Construct package name
        package_name = "bioio_czi/test_resources"

        # Check for dry run
        if args.dry_run:
            # Attempt to build the package
            top_hash = package.build(package_name)

            # Get resolved save path
            manifest_save_path = Path("upload_manifest.jsonl").resolve()
            with open(manifest_save_path, "w") as manifest_write:
                package.dump(manifest_write)

            # Report where manifest was saved
            log.info(f"Dry run generated manifest stored to: {manifest_save_path}")
            log.info(f"Completed package dry run. Result hash: {top_hash}")

        # Upload
        else:
            # Check pre-approved push
            if args.preapproved:
                confirmation = True
            else:
                # Get upload confirmation
                confirmation = None
                while confirmation is None:
                    # Get user input
                    user_input = input("Upload [y]/n? ")

                    # If the user simply pressed enter assume yes
                    if len(user_input) == 0:
                        user_input = "y"
                    # Get first character and lowercase
                    else: 
                        user_input = user_input[0].lower()

                        # Set confirmation from None to a value
                        if user_input == "y":
                            confirmation = True
                        elif user_input == "n":
                            confirmation = False

            # Check confirmation
            if confirmation:
                pushed = package.push(
                    package_name,
                    "s3://bioio-dev-test-resources",
                    message=f"Test resources for `bioio_czi` version: {__version__}.",
                )

                log.info(f"Completed package push. Result hash: {pushed.top_hash}")

                # Update hash file
                with open(Path(__file__).parent / "TEST_RESOURCES_HASH.txt", "w") as hash_file:
                    hash_file.write(pushed.top_hash)
            else:
                log.info(f"Upload canceled.")

    # Catch any exception
    except Exception as e:
        log.error("=============================================")
        if args.debug:
            log.error("\n\n" + traceback.format_exc())
            log.error("=============================================")
        log.error("\n\n" + str(e) + "\n")
        log.error("=============================================")
        sys.exit(1)


###############################################################################
# Runner


def main():
    args = Args()
    upload_test_resources(args)


###############################################################################
# Allow caller to directly run this module (usually in development scenarios)

if __name__ == "__main__":
    main()


from typing import Dict, Tuple

# Example bounding box: {'X': (0, 100), 'Y': (0, 100), 'Z': (0, 20)}
BoundingBox = Dict[str, Tuple[int, int]]


def size(bounding_box: BoundingBox, dim: str) -> int:
    """
    Return the size of the dimension if it is in the bounding box, otherwise -1.
    """
    if dim not in bounding_box:
        return -1
    bounds = bounding_box[dim]
    return bounds[1] - bounds[0]



from typing import Optional
from xml.etree import ElementTree as ET

from bioio_base.types import PhysicalPixelSizes

from .metadata import UnsupportedMetadataError


def get_physical_pixel_sizes(metadata: ET.Element) -> PhysicalPixelSizes:
    """
    Look up pixel sizes for X, Y, and Z dimensions from a CZI's XML.
    """
    return PhysicalPixelSizes(
        Z=_single_physical_pixel_size(metadata, "Z", allow_none=True),
        Y=_single_physical_pixel_size(metadata, "Y"),
        X=_single_physical_pixel_size(metadata, "X"),
    )


def _single_physical_pixel_size(
    metadata: ET.Element, dimension: str, allow_none: bool = False
) -> Optional[float]:
    """
    Look up physical pixel size for one dimension.
    """
    scales = metadata.findall(f"./Metadata/Scaling/Items/Distance[@Id='{dimension}']")

    if len(scales) != 1:
        if allow_none and len(scales) == 0:
            return None
        raise UnsupportedMetadataError(
            f"Expected 1 distance scale for dimension '{dimension}' but found "
            f"{len(scales)}."
        )

    unparsed_scale = scales[0].find("./Value")
    if unparsed_scale is None or unparsed_scale.text is None:
        raise UnsupportedMetadataError(
            f"Could not find any distance scale for dimension '{dimension}'."
        )
    scale = float(unparsed_scale.text)
    # The values are stored in units of meters always in .czi. Convert to
    # microns.
    return scale / 1e-6



import logging
import xml.etree.ElementTree as ET
from copy import copy
from datetime import timedelta
from pathlib import Path
from typing import Any, Dict, Hashable, List, Optional, Tuple, Union

import dask.array as da
import numpy as np
import xarray as xr
from _aicspylibczi import BBox, TileInfo
from aicspylibczi import CziFile
from bioio_base import constants, exceptions
from bioio_base import io as io_utils
from bioio_base import types
from bioio_base.dimensions import (
    DEFAULT_CHUNK_DIMS,
    REQUIRED_CHUNK_DIMS,
    DimensionNames,
    Dimensions,
)
from bioio_base.reader import Reader as BaseReader
from dask import delayed
from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem

from .. import metadata as metadata_utils
from ..bounding_box import size
from ..channels import get_channel_names
from ..pixel_sizes import get_physical_pixel_sizes
from .subblock_metadata import time_between_subblocks

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

CZI_SAMPLES_DIM_CHAR = "A"
CZI_BLOCK_DIM_CHAR = "B"
CZI_SCENE_DIM_CHAR = "S"


###############################################################################

PIXEL_DICT = {
    "gray8": np.uint8,
    "gray16": np.uint16,
    "gray32": np.uint32,
    "gray32float": np.float32,
    "bgr24": np.uint8,
    "bgr48": np.uint16,
    "invalid": np.uint8,
}


class Reader(BaseReader):
    """
    Wraps the aicspylibczi API to provide the same BioIO Reader plugin for
    volumetric Zeiss CZI images.

    Notes
    -----
    To use this reader, install with: `pip install aicspylibczi>=3.1.1`.
    """

    _xarray_dask_data: Optional["xr.DataArray"] = None
    _xarray_data: Optional["xr.DataArray"] = None
    _mosaic_xarray_dask_data: Optional["xr.DataArray"] = None
    _mosaic_xarray_data: Optional["xr.DataArray"] = None
    _dims: Optional[Dimensions] = None
    _metadata: Optional[Any] = None
    _scenes: Optional[Tuple[str, ...]] = None
    _current_scene_index: int = 0
    # Do not provide default value because
    # they may not need to be used by your reader (i.e. input param is an array)
    _fs: "AbstractFileSystem"
    _path: str

    @staticmethod
    def _is_supported_image(fs: AbstractFileSystem, path: str, **kwargs: Any) -> bool:
        if not isinstance(fs, LocalFileSystem):
            raise exceptions.UnsupportedFileFormatError(
                "bioio-czi[aicspylibczi mode]",
                path,
                "Try not setting use_aicspylibczi?",
            )
        try:
            with fs.open(path) as open_resource:
                CziFile(open_resource.f)
                return True
        except RuntimeError as e:
            raise exceptions.UnsupportedFileFormatError(
                "bioio-czi[aicspylibczi mode]",
                path,
                str(e),
            )

    def __init__(
        self,
        image: types.PathLike,
        chunk_dims: Union[str, List[str]] = DEFAULT_CHUNK_DIMS,
        include_subblock_metadata: bool = False,
        fs_kwargs: Dict[str, Any] = {},
    ):
        """
        Parameters
        ----------
        image: types.PathLike
            Path to image file to construct Reader for.
        chunk_dims: Union[str, List[str]]
            Which dimensions to create chunks for.
            Default: DEFAULT_CHUNK_DIMS
            Note: DimensionNames.SpatialY, DimensionNames.SpatialX, and
            DimensionNames.Samples, will always be added to the list if not present
            during dask array construction.
        include_subblock_metadata: bool
            Whether to append metadata from the subblocks to the rest of the embeded
            metadata.
        fs_kwargs: Dict[str, Any]
            Any specific keyword arguments to pass to the fsspec-created filesystem.
            Default: {}
        """
        # Expand details of provided image
        self._fs, self._path = io_utils.pathlike_to_fs(
            image,
            enforce_exists=True,
            fs_kwargs=fs_kwargs,
        )

        # Catch non-local file system
        if not isinstance(self._fs, LocalFileSystem):
            raise exceptions.UnsupportedFileFormatError(
                "bioio-czi[aicspylibczi mode]",
                self._path,
                "Try not setting use_aicspylibczi?",
            )

        # Store params
        if isinstance(chunk_dims, str):
            chunk_dims = list(chunk_dims)

        self.chunk_dims = chunk_dims

        self._include_subblock_metadata = include_subblock_metadata

        # Delayed storage
        self._px_sizes: Optional[types.PhysicalPixelSizes] = None
        self._mapped_dims: Optional[str] = None

        # Enforce valid image
        if not self._is_supported_image(self._fs, self._path):
            raise exceptions.UnsupportedFileFormatError(
                self.__class__.__name__, self._path
            )

    @property
    def mapped_dims(self) -> str:
        if self._mapped_dims is None:
            with self._fs.open(self._path) as open_resource:
                czi = CziFile(open_resource.f)
                self._mapped_dims = Reader._fix_czi_dims(czi.dims)

        return self._mapped_dims

    def _reset_self(self) -> None:
        super()._reset_self()
        self._mapped_dims = None
        self._px_sizes = None

    @staticmethod
    def _fix_czi_dims(dims: str) -> str:
        return (
            dims.replace(CZI_BLOCK_DIM_CHAR, "")
            .replace(CZI_SCENE_DIM_CHAR, "")
            .replace(CZI_SAMPLES_DIM_CHAR, DimensionNames.Samples)
        )

    @property
    def scenes(self) -> Tuple[str, ...]:
        """Note: scenes with no name (`None`) will be renamed to
        "filename-<scene index>" to prevent ambiguity. Similarly, scenes with same
        names are automatically appended with occurrence number to distinguish
        between the two.

        Returns:
            Tuple[str, ...]: Scene names/id
        """
        if self._scenes is None:
            with self._fs.open(self._path) as open_resource:
                czi = CziFile(open_resource.f)
                xpath_str = "./Metadata/Information/Image/Dimensions/S/Scenes/Scene"
                meta_scenes = czi.meta.findall(xpath_str)
                scene_names: List[str] = []

                # mapping of scene name to occurrences, indicating duplication.
                scene_name_frequency = {}
                for scene_idx, meta_scene in enumerate(meta_scenes):
                    shape = meta_scene.find("Shape")
                    if shape is not None:
                        shape_name = shape.get("Name")
                        scene_name = meta_scene.get("Name")
                        combined_scene_name = f"{scene_name}-{shape_name}"
                    else:
                        combined_scene_name = meta_scene.get("Name")
                        # Some scene names can be unpopulated, for those we should fill
                        # with filename-idx
                        if combined_scene_name is None:
                            fname_prefix = Path(self._path).stem
                            combined_scene_name = f"{fname_prefix}-{scene_idx}"
                        # Check for duplicated names
                        # first encounter with a duplicate modify original scene name
                        # to reflect its new duplicate status
                        if combined_scene_name not in scene_name_frequency:
                            scene_name_frequency[combined_scene_name] = [scene_idx, 1]
                        else:
                            if scene_name_frequency[combined_scene_name][1] == 1:
                                scene_names[
                                    scene_name_frequency[combined_scene_name][0]
                                ] += "-1"

                            scene_name_frequency[combined_scene_name][1] += 1

                            combined_scene_name += (
                                f"-{scene_name_frequency[combined_scene_name][1]}"
                            )

                    scene_names.append(combined_scene_name)

                # If the scene is implicit just assign it name Scene:0
                if len(scene_names) < 1:
                    scene_names = [metadata_utils.generate_ome_image_id(0)]
                else:
                    # reconcile scene list against the dims shape
                    dims_shape = czi.get_dims_shape()
                    if len(scene_names) != len(dims_shape) and czi.shape_is_consistent:
                        dims_shape_dict = dims_shape[0]
                        scene_range = dims_shape_dict.get(CZI_SCENE_DIM_CHAR)
                        if scene_range is not None:
                            scene_names = scene_names[scene_range[0] : scene_range[1]]
                        else:
                            # If this is the root node of a split multiscene czi,
                            # then the scene_range could be None because the dims_shape
                            # will be effectively empty.
                            # We do not currently support loading multi-file split
                            # scene CZI files
                            log.warning(
                                "CZI file appears to contain multiple scenes but "
                                "dimension data is not available in this file. "
                                "Root node of split multi-scene CZI files are not "
                                "supported by Reader."
                            )

                self._scenes = tuple(scene_names)

        return self._scenes

    @staticmethod
    def _dims_shape_to_scene_dims_shape(
        dims_shape: List[Dict], scene_index: int, consistent: bool
    ) -> Dict[str, Tuple[int, int]]:
        """
        This function takes the output of `get_dims_shape()` and returns a
        dictionary of dimensions for the selected scene

        Parameters
        ----------
        dims_shape: List[Dict]
            a list of dictionaries, generated by `get_dims_shape()`
        scene_index: int
            the index of the scene being used
        consistent: bool
            true if the dictionaries are consistent could be represented
            compactly (dims_shape with length 1)

        Returns
        -------
        A dictionary of dimensions, ie
        {"T": (0, 1), "C": (0, 3), "Y": (0, 256), "X":(0, 256)}.
        """
        dims_shape_index = 0 if consistent else scene_index
        dims_shape_dict = dims_shape[dims_shape_index]
        dims_shape_dict.pop(CZI_SCENE_DIM_CHAR, None)
        return dims_shape_dict

    @staticmethod
    def _adjust_scene_index(
        dims_shape: List[Dict], scene_index: int, consistent: bool
    ) -> int:
        """
        This function modifies a scene index to be an offset into the true scene
        indices reported by the czi file

        Parameters
        ----------
        dims_shape: List[Dict]
            a list of dictionaries, generated by `get_dims_shape()`
        scene_index: int
            the index of the scene being used
        consistent: bool
            true if the dictionaries are consistent could be represented
            compactly (dims_shape with length 1)

        Returns
        -------
        An int representing the scene index to use in a true libCZI dimension
        """
        dims_shape_index = 0 if consistent else scene_index
        dims_shape_dict = dims_shape[dims_shape_index]
        scene_range = dims_shape_dict.get(CZI_SCENE_DIM_CHAR)
        if scene_range is None:
            return scene_index
        if not consistent:
            # we have selected a dims_shape_dict already based on scene index
            # let's make sure the scene index is in the S range
            if scene_index < scene_range[0] or scene_index >= scene_range[1]:
                raise ValueError(
                    f"Scene index {scene_index} is not in the range "
                    f"{scene_range[0]} to {scene_range[1]}"
                )
            return scene_index
        return scene_range[0] + scene_index

    @staticmethod
    def _read_chunk_from_image(
        fs: AbstractFileSystem,
        path: str,
        scene: int,
        read_dims: Optional[Dict[str, int]] = None,
    ) -> np.ndarray:
        return Reader._get_image_data(
            fs=fs, path=path, scene=scene, read_dims=read_dims
        )[0]

    @staticmethod
    def _get_image_data(
        fs: AbstractFileSystem,
        path: str,
        scene: int,
        read_dims: Optional[Dict[str, int]] = None,
    ) -> Tuple[np.ndarray, List[Tuple[str, int]]]:
        """
        Read and return the squeezed image data requested along with the dimension info
        that was read.

        Parameters
        ----------
        fs: AbstractFileSystem
            The file system to use for reading.
        path: str
            The path to the file to read.
        scene: int
            The scene index to pull the chunk from.
        read_dims: Optional[Dict[str, int]]
            The dimensions to read from the file as a dictionary of string to integer.
            Default: None (Read all data from the image)

        Returns
        -------
        chunk: np.ndarray
            The image chunk read as a numpy array.
        read_dimensions: List[Tuple[str, int]]]
            The dimension sizes that were returned from the read.
        """
        # Catch optional read dim
        if read_dims is None:
            read_dims = {}

        # Init czi
        with fs.open(path) as open_resource:
            czi = CziFile(open_resource.f)

            # Get current scene read dims
            adjusted_scene_index = Reader._adjust_scene_index(
                czi.get_dims_shape(), scene, czi.shape_is_consistent
            )
            read_dims[CZI_SCENE_DIM_CHAR] = adjusted_scene_index

            # Read image
            data, dims = czi.read_image(**read_dims)

            # Drop dims that shouldn't be provided back
            ops: List[Union[int, slice]] = []
            real_dims = []
            for dim_info in dims:
                # Expand dimension info
                dim, _ = dim_info

                # If the dim was provided in the read dims
                # we know a single plane for that dimension was requested so remove it
                if dim in read_dims or dim == CZI_BLOCK_DIM_CHAR:
                    ops.append(0)

                # Otherwise just read the full slice
                else:
                    ops.append(slice(None, None, None))
                    real_dims.append(dim_info)

            # Convert ops and run getitem
            return data[tuple(ops)], real_dims

    def _create_dask_array(self, czi: CziFile) -> xr.DataArray:
        """
        Creates a delayed dask array for the file.

        Parameters
        ----------
        czi: CziFile
            An open CziFile for processing.

        Returns
        -------
        image_data: da.Array
            The fully constructed and fully delayed image as a Dask Array object.
        """
        # Always add the plane dimensions if not present already
        for dim in REQUIRED_CHUNK_DIMS:
            if dim not in self.chunk_dims:
                self.chunk_dims.append(dim)

        # Safety measure / "feature"
        self.chunk_dims = [d.upper() for d in self.chunk_dims]

        # Construct the delayed dask array
        dims_shape = Reader._dims_shape_to_scene_dims_shape(
            czi.get_dims_shape(),
            scene_index=self.current_scene_index,
            consistent=czi.shape_is_consistent,
        )

        # Remove block dim as not useful
        dims_shape.pop(CZI_BLOCK_DIM_CHAR, None)

        dims_str = czi.dims
        for remove_dim_char in [CZI_BLOCK_DIM_CHAR, CZI_SCENE_DIM_CHAR]:
            dims_str = dims_str.replace(remove_dim_char, "")

        # Get the shape for the chunk and operating shape for the dask array
        # We also collect the chunk and non chunk dimension ordering so that we can
        # swap the dimensions after we
        # block the dask array together.
        sample_chunk_shape = []
        operating_shape = []
        non_chunk_dimension_ordering = []
        chunk_dimension_ordering = []
        for i, dim in enumerate(dims_str):
            # Unpack dim info
            _, dim_size = dims_shape[dim]

            # If the dim is part of the specified chunk dims then append it to the
            # sample, and, append the dimension
            # to the chunk dimension ordering
            if dim in self.chunk_dims:
                sample_chunk_shape.append(dim_size)
                chunk_dimension_ordering.append(dim)

            # Otherwise, append the dimension to the non chunk dimension ordering, and,
            # append the true size of the
            # image at that dimension
            else:
                non_chunk_dimension_ordering.append(dim)
                operating_shape.append(dim_size)

        # Convert shapes to tuples and combine the non and chunked dimension orders as
        # that is the order the data will
        # actually come out of the read data as
        sample_chunk_shape_tuple = tuple(sample_chunk_shape)
        blocked_dimension_order = (
            non_chunk_dimension_ordering + chunk_dimension_ordering
        )

        # Fill out the rest of the operating shape with dimension sizes of 1 to match
        # the length of the sample chunk
        # When dask.block happens it fills the dimensions from inner-most to outer-most
        # with the chunks as long as the dimension is size 1
        # Basically, we are adding empty dimensions to the operating shape that will be
        # filled by the chunks from dask
        operating_shape_tuple = tuple(operating_shape) + (1,) * len(
            sample_chunk_shape_tuple
        )

        # Create empty numpy array with the operating shape so that we can iter through
        # and use the multi_index to create the readers.
        lazy_arrays: np.ndarray = np.ndarray(operating_shape_tuple, dtype=object)

        # We can enumerate over the multi-indexed array and construct read_dims
        # dictionaries by simply zipping together the ordered dims list and the current
        # multi-index plus the begin index for that plane. We then set the value of the
        # array at the same multi-index to the delayed reader using the constructed
        # read_dims dictionary.
        dims = [
            d for d in czi.dims if d not in [CZI_BLOCK_DIM_CHAR, CZI_SCENE_DIM_CHAR]
        ]
        begin_indicies = tuple(dims_shape[d][0] for d in dims)
        for np_index, _ in np.ndenumerate(lazy_arrays):
            # Add the czi file begin index for each dimension to the array dimension
            # index
            this_chunk_read_indicies = (
                current_dim_begin_index + curr_dim_index
                for current_dim_begin_index, curr_dim_index in zip(
                    begin_indicies, np_index
                )
            )

            # Zip the dims with the read indices
            this_chunk_read_dims = dict(
                zip(blocked_dimension_order, this_chunk_read_indicies)
            )

            # Remove the dimensions that we want to chunk by from the read dims
            for d in self.chunk_dims:
                this_chunk_read_dims.pop(d, None)

            # Get pixel type and catch unsupported
            pixel_type = PIXEL_DICT.get(czi.pixel_type)
            if pixel_type is None:
                raise TypeError(f"Pixel type: {czi.pixel_type} is not supported.")

            # Add delayed array to lazy arrays at index
            lazy_arrays[np_index] = da.from_delayed(
                delayed(Reader._read_chunk_from_image)(
                    fs=self._fs,
                    path=self._path,
                    scene=self.current_scene_index,
                    read_dims=this_chunk_read_dims,
                ),
                shape=sample_chunk_shape,
                dtype=pixel_type,
            )

        # Convert the numpy array of lazy readers into a dask array and fill the inner
        # most empty dimensions with chunks
        merged = da.block(lazy_arrays.tolist())

        # Because we have set certain dimensions to be chunked and others not
        # we will need to transpose back to original dimension ordering
        # Example being, if the original dimension ordering was "SZYX" and we want to
        # chunk by "S", "Y", and "X" we created an array with dimensions ordering "ZSYX"
        transpose_indices = []
        transpose_required = False
        for i, d in enumerate(dims_str):
            new_index = blocked_dimension_order.index(d)
            if new_index != i:
                transpose_required = True
                transpose_indices.append(new_index)
            else:
                transpose_indices.append(i)

        # Only run if the transpose is actually required
        # The default case is "Z", "Y", "X", which _usually_ doesn't need to be
        # transposed because that is _usually_ the normal dimension order of the CZI
        # file anyway
        if transpose_required:
            merged = da.transpose(merged, tuple(transpose_indices))

        # Because dimensions outside of Y and X can be in any order and present or not
        # we also return the dimension order string.
        return merged

    @staticmethod
    def _get_coords_and_physical_px_sizes(
        xml: ET.Element, scene_index: int, dims_shape: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], types.PhysicalPixelSizes]:
        # Create coord dict
        coords: Dict[str, Any] = {}

        # Attach channel names to coords
        scene_channel_list = get_channel_names(xml, scene_index, dims_shape)
        if scene_channel_list is not None:
            coords[DimensionNames.Channel] = scene_channel_list

        # Handle Spatial Dimensions
        px_sizes = get_physical_pixel_sizes(xml)
        for dim_name, scale in px_sizes._asdict().items():
            if scale is not None and dim_name in dims_shape:
                dim_size = size(dims_shape, dim_name)
                coords[dim_name] = Reader._generate_coord_array(0, dim_size, scale)

        return coords, px_sizes

    def _read_delayed(self) -> xr.DataArray:
        """
        Construct the delayed xarray DataArray object for the image.

        Returns
        -------
        image: xr.DataArray
            The fully constructed and fully delayed image as a DataArray object.
            Metadata is attached in some cases as coords, dims, and attrs.

        Raises
        ------
        exceptions.UnsupportedFileFormatError
            The file could not be read or is not supported.
        """
        with self._fs.open(self._path) as open_resource:
            czi = CziFile(open_resource.f)

            dims_shape = Reader._dims_shape_to_scene_dims_shape(
                dims_shape=czi.get_dims_shape(),
                scene_index=self.current_scene_index,
                consistent=czi.shape_is_consistent,
            )

            # Get dims as list for xarray
            img_dims_list = list(self.mapped_dims)

            # Get image data
            image_data = self._create_dask_array(czi)

            # Create coordinate planes
            meta = czi.meta
            coords, px_sizes = self._get_coords_and_physical_px_sizes(
                xml=meta,
                scene_index=self.current_scene_index,
                dims_shape=dims_shape,
            )

            # Append subblock metadata to the other metadata if param is True
            if self._include_subblock_metadata:
                subblocks = czi.read_subblock_metadata(unified_xml=True)
                meta.append(subblocks)

            # Store pixel sizes
            self._px_sizes = px_sizes

            # handle edge case where image has 0,0 YX dims:
            if image_data.shape[-2:] == (0, 0):
                return xr.DataArray(
                    dims=coords.keys(),
                    coords=coords,
                    attrs={constants.METADATA_UNPROCESSED: meta},
                )
            else:
                return xr.DataArray(
                    image_data,
                    dims=img_dims_list,
                    coords=coords,
                    attrs={constants.METADATA_UNPROCESSED: meta},
                )

    def _read_immediate(self) -> xr.DataArray:
        """
        Construct the in-memory xarray DataArray object for the image.

        Returns
        -------
        image: xr.DataArray
            The fully constructed and fully read into memory image as a DataArray
            object. Metadata is attached in some cases as coords, dims, and attrs.

        Raises
        ------
        exceptions.UnsupportedFileFormatError
            The file could not be read or is not supported.
        """
        with self._fs.open(self._path) as open_resource:
            czi = CziFile(open_resource.f)
            dims_shape = Reader._dims_shape_to_scene_dims_shape(
                dims_shape=czi.get_dims_shape(),
                scene_index=self.current_scene_index,
                consistent=czi.shape_is_consistent,
            )

            # Get image data
            image_data, _ = self._get_image_data(
                fs=self._fs,
                path=self._path,
                scene=self.current_scene_index,
            )

            # Get metadata
            meta = czi.meta

            # Create coordinate planes
            coords, px_sizes = self._get_coords_and_physical_px_sizes(
                xml=meta,
                scene_index=self.current_scene_index,
                dims_shape=dims_shape,
            )

            # Store pixel sizes
            self._px_sizes = px_sizes

            return xr.DataArray(
                image_data,
                dims=[d for d in self.mapped_dims],
                coords=coords,
                attrs={constants.METADATA_UNPROCESSED: meta},
            )

    @staticmethod
    def _stitch_tiles(
        data: types.ArrayLike,
        data_dims: str,
        data_dims_shape: Dict[str, Tuple[int, int]],
        tile_bboxes: Dict[TileInfo, BBox],
        final_bbox: BBox,
    ) -> types.ArrayLike:
        # Assumptions: 1) docs for ZEISSRAW(CZI) say:
        #   Scene – for clustering items in X/Y direction (data belonging to
        #   contiguous regions of interests in a mosaic image).

        # Store the mosaic array shape
        arr_shape_list = []

        ordered_dims_present = [
            dim
            for dim in data_dims
            if dim not in [CZI_BLOCK_DIM_CHAR, DimensionNames.MosaicTile]
        ]
        for dim in ordered_dims_present:
            if dim not in REQUIRED_CHUNK_DIMS:
                arr_shape_list.append(data_dims_shape[dim][1])
            if dim is DimensionNames.SpatialY:
                arr_shape_list.append(final_bbox.h)
            if dim is DimensionNames.SpatialX:
                arr_shape_list.append(final_bbox.w)
            if dim is DimensionNames.Samples:
                arr_shape_list.append(data_dims_shape[CZI_SAMPLES_DIM_CHAR][1])

        ans = None
        if isinstance(data, da.Array):
            ans = da.zeros(
                shape=tuple(arr_shape_list),
                dtype=data.dtype,
            )
        else:
            ans = np.zeros(arr_shape_list, dtype=data.dtype)

        for tile_info, box in tile_bboxes.items():
            # Construct data indexes to use
            tile_dims = tile_info.dimension_coordinates
            tile_dims.pop(CZI_SCENE_DIM_CHAR, None)
            tile_dims.pop(CZI_BLOCK_DIM_CHAR, None)
            data_indexes = [
                tile_dims[t_dim]
                for t_dim in data_dims
                if t_dim not in REQUIRED_CHUNK_DIMS
            ]
            # Add Y and X
            data_indexes.append(slice(None))  # Y ":"
            data_indexes.append(slice(None))  # X ":"
            if CZI_SAMPLES_DIM_CHAR in tile_dims.keys():
                data_indexes.append(slice(None))

            # Construct data indexes for ans
            ans_indexes = []
            for dim in ordered_dims_present:
                if dim not in [
                    DimensionNames.MosaicTile,
                    DimensionNames.Samples,
                    DimensionNames.SpatialY,
                    DimensionNames.SpatialX,
                ]:
                    if dim in tile_dims.keys():
                        ans_indexes.append(tile_dims[dim])

                if dim is DimensionNames.SpatialY:
                    start = box.y - final_bbox.y
                    ans_indexes.append(slice(start, start + box.h, 1))
                if dim is DimensionNames.SpatialX:
                    start = box.x - final_bbox.x
                    ans_indexes.append(slice(start, start + box.w, 1))
                if dim is DimensionNames.Samples:
                    ans_indexes.append(slice(None))

            # Assign the tiles into ans
            ans[tuple(ans_indexes)] = data[tuple(data_indexes)]

        return ans

    def _construct_mosaic_xarray(self, data: types.ArrayLike) -> xr.DataArray:
        # Get max of mosaic positions from lif
        with self._fs.open(self._path) as open_resource:
            czi = CziFile(open_resource.f)
            dims_shape = Reader._dims_shape_to_scene_dims_shape(
                dims_shape=czi.get_dims_shape(),
                scene_index=self.current_scene_index,
                consistent=czi.shape_is_consistent,
            )

            bboxes = czi.get_all_mosaic_tile_bounding_boxes(S=self.current_scene_index)
            mosaic_scene_bbox = czi.get_mosaic_scene_bounding_box(
                index=self.current_scene_index
            )

            # Stitch
            stitched = self._stitch_tiles(
                data=data,
                data_dims=self.mapped_dims,
                data_dims_shape=dims_shape,
                tile_bboxes=bboxes,
                final_bbox=mosaic_scene_bbox,
            )

            # Copy metadata
            dims = [
                d
                for d in self.xarray_dask_data.dims
                if d is not DimensionNames.MosaicTile
            ]
            coords: Dict[Hashable, Any] = {
                d: v
                for d, v in self.xarray_dask_data.coords.items()
                if d
                not in [
                    DimensionNames.MosaicTile,
                    DimensionNames.SpatialY,
                    DimensionNames.SpatialX,
                ]
            }

            # Add expanded Y and X coords
            if self.physical_pixel_sizes.Y is not None:
                dim_y_index = dims.index(DimensionNames.SpatialY)
                coords[DimensionNames.SpatialY] = Reader._generate_coord_array(
                    0, stitched.shape[dim_y_index], self.physical_pixel_sizes.Y
                )
            if self.physical_pixel_sizes.X is not None:
                dim_x_index = dims.index(DimensionNames.SpatialX)
                coords[DimensionNames.SpatialX] = Reader._generate_coord_array(
                    0, stitched.shape[dim_x_index], self.physical_pixel_sizes.X
                )

            attrs = copy(self.xarray_dask_data.attrs)

            return xr.DataArray(
                data=stitched,
                dims=dims,
                coords=coords,
                attrs=attrs,
            )

    def _get_stitched_dask_mosaic(self) -> xr.DataArray:
        return self._construct_mosaic_xarray(self.dask_data)

    def _get_stitched_mosaic(self) -> xr.DataArray:
        return self._construct_mosaic_xarray(self.data)

    @property
    def physical_pixel_sizes(self) -> types.PhysicalPixelSizes:
        """
        Returns
        -------
        sizes: PhysicalPixelSizes
            Using available metadata, the floats representing physical pixel sizes for
            dimensions Z, Y, and X.

        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        if self._px_sizes is None:
            # We get pixel sizes as a part of array construct
            # so simply run array construct
            self.dask_data

        if self._px_sizes is None:
            raise ValueError("Pixel sizes weren't created as a part of image reading")

        return self._px_sizes

    def get_mosaic_tile_position(
        self,
        mosaic_tile_index: int,
        **kwargs: int,
    ) -> Tuple[int, int]:
        """
        Get the absolute position of the top left point for a single mosaic tile.

        Parameters
        ----------
        mosaic_tile_index: int
            The index for the mosaic tile to retrieve position information for.
        kwargs: int
            The keywords below allow you to specify the dimensions that you wish
            to match. If you under-specify the constraints you can easily
            end up with a massive image stack.
                       Z = 1   # The Z-dimension.
                       C = 2   # The C-dimension ("channel").
                       T = 3   # The T-dimension ("time").

        Returns
        -------
        top: int
            The Y coordinate for the tile position.
        left: int
            The X coordinate for the tile position.

        Raises
        ------
        UnexpectedShapeError
            The image has no mosaic dimension available.

        Notes
        -----
        Defaults T and C dimensions to 0 if present as dimensions in image
        to avoid reading in massive image stack for large files.
        """
        if DimensionNames.MosaicTile not in self.dims.order:
            raise exceptions.UnexpectedShapeError("No mosaic dimension in image.")

        with self._fs.open(self._path) as open_resource:
            czi = CziFile(open_resource.f)

            # Default Channel and Time dimensions to 0 to improve
            # worst case read time for large files **only**
            # when those dimensions are present on the image.
            for dimension_name in [DimensionNames.Channel, DimensionNames.Time]:
                if dimension_name not in kwargs and dimension_name in self.dims.order:
                    kwargs[dimension_name] = 0

            bbox = czi.get_mosaic_tile_bounding_box(
                M=mosaic_tile_index, S=self.current_scene_index, **kwargs
            )
            return bbox.y, bbox.x

    def get_mosaic_tile_positions(self, **kwargs: int) -> List[Tuple[int, int]]:
        """
        Get the absolute positions of the top left points for each mosaic tile
        matching the specified dimensions and current scene.

        Parameters
        ----------
        kwargs: int
            The keywords below allow you to specify the dimensions that you wish
            to match. If you under-specify the constraints you can easily
            end up with a massive image stack.
                       Z = 1   # The Z-dimension.
                       C = 2   # The C-dimension ("channel").
                       T = 3   # The T-dimension ("time").

        Returns
        -------
        mosaic_tile_positions: List[Tuple[int, int]]
            List of the Y and X coordinate for the tile positions.

        Raises
        ------
        UnexpectedShapeError
            The image has no mosaic dimension available.
        """
        if DimensionNames.MosaicTile not in self.dims.order:
            raise exceptions.UnexpectedShapeError("No mosaic dimension in image.")

        with self._fs.open(self._path) as open_resource:
            czi = CziFile(open_resource.f)

            tile_info_to_bboxes = czi.get_all_mosaic_tile_bounding_boxes(
                S=self.current_scene_index, **kwargs
            )

            # Convert dictionary of tile info mappings to
            # a list of bounding boxes sorted according to their
            # respective M indexes
            m_indexes_to_mosaic_positions = {
                tile_info.m_index: (bbox.y, bbox.x)
                for tile_info, bbox in tile_info_to_bboxes.items()
            }
            return [
                m_indexes_to_mosaic_positions[m_index]
                for m_index in sorted(m_indexes_to_mosaic_positions.keys())
            ]

    @property
    def time_interval(self) -> Optional[timedelta]:
        """
        Extracts the the average time interval between consecutive timepoints
        as a timedelta object.

        Returns
        -------
        Optional[timedelta]
            Average interval between timepoints.
            Returns None if total_time_duration is None or less than two timepoints.
        """

        # The purpose of this conditional is to not log a warning for files with a
        # single timepoint.
        timepoints = (
            self.dims[DimensionNames.Time][0]
            if DimensionNames.Time in self.dims.order
            else None
        )
        if timepoints is None or timepoints < 2:
            return None

        total_duration = self.total_time_duration
        if total_duration is None:
            return None
        return total_duration / (timepoints - 1)

    @property
    def total_time_duration(self) -> Optional[timedelta]:
        """
        Extracts the total duration of the timelapse as a timedelta object.
        This is the time between the first acquisition and the first acquisition of the
        last timepoint.

        Returns
        -------
        Optional[timedelta]
            Total time duration as a timedelta object.
            Returns None if extraction fails.
        """
        timepoints = (
            self.dims[DimensionNames.Time][0]
            if DimensionNames.Time in self.dims.order
            else None
        )
        if timepoints is None or timepoints < 2:
            return None

        try:
            with self._fs.open(self._path) as open_resource:
                czi = CziFile(open_resource.f)
                duration_ms = time_between_subblocks(
                    czi,
                    self.current_scene_index,
                    start_frame=0,
                    # Index of the last timepoint is one less than the number of
                    # timepoints
                    end_frame=timepoints - 1,
                )
                return (
                    timedelta(milliseconds=duration_ms)
                    if duration_ms is not None
                    else None
                )

        except Exception as exc:
            log.warning("Failed to extract Total Time Duration: %s", exc, exc_info=True)

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from typing import Any, Callable, ContextManager, Dict, Optional, Tuple, Union
from xml.etree import ElementTree as ET

import dask.array as da
import numpy as np
import xarray as xr
from bioio_base import constants, exceptions, types
from bioio_base.dimensions import (
    DEFAULT_DIMENSION_ORDER_LIST,
    DimensionNames,
    Dimensions,
)
from bioio_base.reader import Reader as BaseReader
from bioio_base.types import PhysicalPixelSizes
from dask import delayed
from fsspec.spec import AbstractFileSystem
from pylibCZIrw import czi

from .. import metadata
from ..channels import get_channel_names, size
from ..metadata import UnsupportedMetadataError
from ..pixel_sizes import get_physical_pixel_sizes

log = logging.getLogger(__name__)

PIXEL_DICT = {
    "Gray8": np.uint8,
    "Gray16": np.uint16,
    "Gray32": np.uint32,  # Not supported by underlying pylibCZIrw
    "Gray32Float": np.float32,
    "Bgr24": np.uint8,
    "Bgr48": np.uint16,
    "Bgr96Float": np.float32,  # Supported by pylibCZIrw but not tested in this plugin
    "invalid": np.uint8,
}


class Reader(BaseReader):
    """
    Wraps the pylibczirw API to provide the same BioIO Reader plugin for
    volumetric Zeiss CZI images.

    Parameters
    ----------
    image: types.PathLike
        Path to image file to construct Reader for.
    fs_kwargs: Dict[str, Any]
        Any specific keyword arguments to pass down to the fsspec created filesystem.
        Default: {}
    """

    _xarray_dask_data: Optional["xr.DataArray"] = None
    _xarray_data: Optional["xr.DataArray"] = None
    _dims: Optional[Dimensions] = None
    _metadata: Optional[ET.Element] = None
    _scenes: Optional[Tuple[str, ...]] = None
    _current_scene_index: int = 0
    _fs: "AbstractFileSystem"
    _path: str

    @staticmethod
    def _is_supported_image(
        fs: AbstractFileSystem,
        path: str,
        **kwargs: Any,
    ) -> bool:
        """
        Check if file is a supported CZI by attempting to open it.

        Parameters
        ----------
        fs: AbstractFileSystem
            The file system to used for reading.
        path: str
            The path to the file to read.
        kwargs: Any
            Any kwargs used for reading and validation of the file.

        Returns
        -------
        supported: bool
            Boolean value indicating if the file is supported by the reader,
            or raises an exception if it is not.
        """
        try:
            with open(path):
                return True
        except RuntimeError as e:
            raise exceptions.UnsupportedFileFormatError(
                "bioio-czi[pylibczirw mode]",
                path,
                str(e),
            )

    def __init__(self, image: types.PathLike, fs_kwargs: Dict[str, Any] = {}) -> None:
        path = str(image)
        try:
            with open(path) as file:
                self._fs = None  # Unused but required by tests
                self._path = path
                self._total_bounding_box = file.total_bounding_box_no_pyramid
                self._pixel_types = file.pixel_types
                self._scenes_bounding_rectangle = (
                    file.scenes_bounding_rectangle_no_pyramid
                )
        except RuntimeError:
            raise exceptions.UnsupportedFileFormatError(self.__class__.__name__, path)

    @property
    def scenes(self) -> Tuple[str, ...]:
        """
        Returns
        -------
        scenes: Tuple[str, ...]
            A tuple of valid scene ids in the file.

        Notes
        -----
        Scene IDs are strings - not a range of integers.

        When iterating over scenes please use:

        >>> for id in image.scenes

        and not:

        >>> for i in range(len(image.scenes))
        """

        def scene_name(metadata: ET.Element, scene_index: int) -> str:
            scene_info = metadata.findall(
                "./Metadata/Information/Image/Dimensions/"
                f"S/Scenes/Scene[@Index='{scene_index}']"
            )
            if len(scene_info) != 1:
                raise UnsupportedMetadataError(
                    f"Expected 1 scene for index '{scene_index}' "
                    "but found {len(scene_info)}."
                )
            scene_name = scene_info[0].get("Name")
            if scene_name is None:
                scene_name = str(scene_index)
            shape_info = scene_info[0].find("Shape")
            if shape_info is not None:
                shape_name = shape_info.get("Name")
                if shape_name is not None:
                    return f"{scene_name}-{shape_name}"
            return scene_name

        if self._scenes is None:
            with open(self._path) as file:
                # Underlying scene IDs are ints
                scene_ids = file.scenes_bounding_rectangle.keys()
                self._scenes = tuple(scene_name(self.metadata, i) for i in scene_ids)
                if len(self._scenes) < 1:
                    # If there are no scenes, use the default scene ID
                    self._scenes = (metadata.generate_ome_image_id(0),)

        return self._scenes

    def _get_coords(
        self, xml: ET.Element, scene_index: int, dims_shape: Dict[str, Any]
    ) -> Dict[str, Union[list, np.ndarray]]:
        """
        Generate coordinate arrays for channel dimension ("C") and spatial dimensions
        ("X", "Y", and "Z") based on channel names and physical pixel sizes.

        Time coordinates are not handled here.
        Hypothetically, we could get the interval between time points from the metadata
        and generate a time coordinate array.
        """
        coords: Dict[str, list | np.ndarray] = {}

        channel_names = get_channel_names(xml, scene_index, dims_shape)
        if channel_names is not None:
            coords[DimensionNames.Channel] = channel_names

        # Handle Spatial Dimensions
        for dim_name, scale in self.physical_pixel_sizes._asdict().items():
            if scale is not None and dim_name in dims_shape:
                dim_size = size(dims_shape, dim_name)
                coords[dim_name] = Reader._generate_coord_array(0, dim_size, scale)

        return coords

    def _array_builder(self, index_dims: list[str]) -> Callable[[tuple[int]], int]:
        """
        Internal helper method to get one chunk of the image data.

        Parameters
        ----------
        index_dims: list[str]
            The names of the dimensions that will be used to select a chunk.

        Returns
        -------
        array_builder: Callable[[tuple[int]], int]
            Function of one parameter indices: tuple[int] that defines the chunk.
            indices must be the same length as index_dims, and in the same order.

        Example
        -------
        >>> self._array_builder(['T', 'C', 'Z'])((0, 1, 2)) = file.read(
        ...   scene=0,
        ...   plane={'T': 0, 'C': 1, 'Z': 2}
        ... )
        """

        def array_builder(indices: tuple[int]) -> int:
            assert len(indices) >= len(
                index_dims
            ), f"Expected {len(indices)} >= {len(index_dims)}."
            # E.g., plane = {'T': 0, 'C': 1, 'Z': 2}
            plane = {d: indices[i] for i, d in enumerate(index_dims)}
            scene: int | None
            if len(self._scenes_bounding_rectangle) == 0:
                # Some files have no scenes but can still be read if scene is not
                # specified.
                scene = None
                roi = None
            else:
                # The purpose of the next 2 lines is complicated.
                # ROI stands for Region Of Interest.
                #
                # In pylibczi's read method, the default ROI is the bounding
                # rectangle of the scene **across all zoom levels**. We are going to
                # read just the highest resolution level (zoom = 1), which is
                # smaller than the default ROI in some cases. For example,
                # scene 0 of the test file S=2_4x2_T=2=Z=3_CH=2.czi is 947x487 when
                # looking at only the highest resolution, but is 948x488 when
                # all zoom levels are considered. (I believe this is because at zoom
                # 0.5, the result is ceiling(947/2) x ceiling(487/2).)
                #
                # See also: file.scenes_bounding_rectangle vs.
                # file.scenes_bounding_rectangle_no_pyramid.
                #
                # Therefore, when calling read, we crop to just the ROI of the
                # highest resolution level.
                scene = self._current_scene_index
                roi = self._scenes_bounding_rectangle[scene]
            with open(self._path) as file:
                result = file.read(scene=scene, plane=plane, roi=roi)
            # result.shape is (Y, X, 1) or (Y, X, 3) depending on whether it's RGB
            # or grayscale. We want to return (Y, X) or (Y, X, 3).
            return np.squeeze(result)

        return array_builder

    def _read_delayed(self) -> xr.DataArray:
        """
        The delayed data array constructor for the image.

        Returns
        -------
        data: xr.DataArray
            The fully constructed delayed DataArray.

            It is additionally recommended to closely monitor how dask array chunks are
            managed.
        """
        # 1. Combine the dimension bounds from total_bounding_box (all dimensions) and
        # scenes_bounding_rectangle (XY only) in order to compute the coordinate array
        # for each dimension. (Think of the coordinate array as the "ticks" on an axis.)
        dim_bounds = self._total_bounding_box
        if len(self._scenes_bounding_rectangle) > 0:
            assert self._current_scene_index in self._scenes_bounding_rectangle, (
                f"Expected {self._current_scene_index} to be in "
                f"{self._scenes_bounding_rectangle}."
            )
            rect = self._scenes_bounding_rectangle[self._current_scene_index]
            dim_bounds[DimensionNames.SpatialX] = (rect.x, rect.x + rect.w)
            dim_bounds[DimensionNames.SpatialY] = (rect.y, rect.y + rect.h)
        coords = self._get_coords(self.metadata, self._current_scene_index, dim_bounds)

        # 2. Figure out which dimensions are available on this image, and put them in
        # TCZYX order as much as possible.
        ordered_dims = [
            d
            for d in DEFAULT_DIMENSION_ORDER_LIST
            if d in coords or size(self._total_bounding_box, d) > 1
        ]
        assert ordered_dims[-2:] == [DimensionNames.SpatialY, DimensionNames.SpatialX]
        # E.g., non_yx_dims = ['T', 'C', 'Z']
        non_yx_dims = ordered_dims[:-2]

        # 4. Determine the chunk sizes and number of chunks. Each chunk is a single
        # YX slice.
        # E.g., shape = (30, 2, 20, 100, 100)
        shape = tuple(
            len(coords[d]) if d in coords else size(self._total_bounding_box, d)
            for d in ordered_dims
        )
        # E.g., shape_without_yx = (30, 2, 20)
        shape_without_yx = shape[:-2]

        chunk_shape = shape[-2:]
        if "Bgr" in self._pixel_types[0]:
            # If the image is BGR, each chunk has shape (X, Y, 3)
            chunk_shape += (3,)
            ordered_dims.append(DimensionNames.Samples)

        # 5. Create delayed chunks
        # The Y and X shape of lazy_arrays are both 1 because we are making each YX
        # slice a single chunk.
        # E.g., lazy_arrays.shape = (30, 2, 20, 1, 1)
        lazy_arrays: np.ndarray = np.ndarray(shape_without_yx + (1, 1), dtype=object)
        for np_index, _ in np.ndenumerate(lazy_arrays):
            lazy_arrays[np_index] = da.from_delayed(
                delayed(self._array_builder(non_yx_dims))(np_index),
                chunk_shape,
                dtype=PIXEL_DICT[self._pixel_types[0]],
            )

        # 6. Package chunks and metadata into a DataArray
        return xr.DataArray(
            data=da.block(lazy_arrays.tolist()),
            dims=ordered_dims,
            coords=coords,
            attrs={constants.METADATA_UNPROCESSED: self.metadata},
        )

    def _read_immediate(self) -> xr.DataArray:
        """
        The immediate data array constructor for the image.

        Returns
        -------
        data: xr.DataArray
            The fully read data array.
        """
        return self._read_delayed().compute()

    def _get_stitched_dask_mosaic(self) -> xr.DataArray:
        """
        This reader always stiches the entire image together, as the underlying
        pylibczirw does not support reading individual tiles.

        Returns
        -------
        mosaic: xr.DataArray
            The fully stitched together image. Contains all the dimensions of the image
            with the YX expanded to the full mosaic.
        """
        return self.xarray_dask_data

    def _get_stitched_mosaic(self) -> xr.DataArray:
        """
        This reader always stiches the entire image together, as the underlying
        pylibczirw does not support reading individual tiles.

        Returns
        -------
        mosaic: np.ndarray
            The fully stitched together image. Contains all the dimensions of the image
            with the YX expanded to the full mosaic.
        """
        return self.xarray_data

    @property
    def mosaic_xarray_dask_data(self) -> xr.DataArray:
        """
        This reader always stiches the entire image together, as the underlying
        pylibczirw does not support reading individual tiles.

        Returns
        -------
        xarray_dask_data: xr.DataArray
            The delayed stiched mosaic image and metadata as an annotated data array.
        """
        return self.xarray_dask_data

    @property
    def mosaic_xarray_data(self) -> xr.DataArray:
        """
        This reader always stiches the entire image together, as the underlying
        pylibczirw does not support reading individual tiles.

        Returns
        -------
        xarray_dask_data: xr.DataArray
            The in-memory stitched mosaic image and metadata as an annotated data array.
        """
        return self.xarray_data

    @property
    def metadata(self) -> ET.Element:
        """
        Returns
        -------
        metadata: Any
            The metadata for the formats supported by the inhereting Reader.

            If the inheriting Reader supports processing the metadata into a more useful
            format / Python object, this will return the result.

            For both the unprocessed and processed metadata from the file, use
            `xarray_dask_data.attrs` which will contain a dictionary with keys:
            `unprocessed` and `processed` that you can then select.

        Caution
        -------
        This method uses the xml.etree.ElementTree.fromstring, which is vulnerable
        to denial of service attacks from malicious input data. To learn more, see:
        https://docs.python.org/3/library/xml.html#xml-vulnerabilities
        """
        if self._metadata is None:
            with open(self._path) as file:
                self._metadata = ET.fromstring(file.raw_metadata)
        return self._metadata

    @property
    def physical_pixel_sizes(self) -> PhysicalPixelSizes:
        """
        Returns
        -------
        sizes: PhysicalPixelSizes
            Using available metadata, the floats representing physical pixel sizes for
            dimensions Z, Y, and X.

        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        return get_physical_pixel_sizes(self.metadata)

    @property
    def mosaic_tile_dims(self) -> None:
        """
        Returns
        -------
        tile_dims: None
            Inherited method. Mosaic tiles are not supported by the underlying
            pylibCZIrw.
        """
        return None

    @property
    def total_time_duration(self) -> None:
        """
        Cannot read time duration accurately without subblock metadata.
        """
        return None


def open(filepath: str) -> ContextManager[czi.CziReader]:
    """
    Wrapper around czi.open_czi to provide type hinting that clarifies the result
    is a czi.CziReader
    """
    if filepath.startswith("http") or filepath.startswith("https"):
        return czi.open_czi(filepath, czi.ReaderFileInputTypes.Curl)
    return czi.open_czi(filepath)



"""
Helper functions extracting metadata from subblocks in aicspylibczi mode.
"""
import logging
import warnings
from typing import Optional

import numpy as np
from aicspylibczi import CziFile
from lxml import etree

log = logging.getLogger(__name__)


def _extract_acquisition_time_from_subblock_metadata(
    subblock_metadata: str,
) -> Optional[np.datetime64]:
    """Extracts acquisition time from subblock metadata."""
    outlxml = etree.fromstring(subblock_metadata)
    acquisition_time_element = outlxml.find(".//AcquisitionTime")

    if acquisition_time_element is not None and acquisition_time_element.text:
        try:
            # Parse acquisition time using numpy's datetime64 because it supports high
            # precision time (sub-microsecond). This parsing treats timezone-less dates
            # as UTC, which is fine for computing durations.
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    "no explicit representation of timezones available",
                    category=UserWarning,
                )
                return np.datetime64(str(acquisition_time_element.text))
        except Exception as exc:
            log.warning("Failed to extract acquisition time: %s", exc, exc_info=True)

    return None


def _acquisition_time(czi: CziFile, scene: int, frame: int) -> Optional[np.datetime64]:
    """Get the time of the first acquisition in the given scene at the given frame."""
    subblocks_at_t: list[tuple[dict, str]] = czi.read_subblock_metadata(
        T=frame, S=scene
    )

    # subblocks_at_t is a list whose element are tuples of the form:
    # ({ Z=0, C=0, T=which_subblock, R=0, S=scene, I=0, H=0, V=0 }, metadata_string)
    acquisition_times = [
        _extract_acquisition_time_from_subblock_metadata(metadata)
        for _, metadata in subblocks_at_t
    ]
    # NaT is numpy's representation of "Not a Time", which is used when parsing fails.
    filtered_acquisition_times = [
        t for t in acquisition_times if t is not None and t is not np.datetime64("NaT")
    ]

    if len(filtered_acquisition_times) == 0:
        return None
    # One timepoint has many acquisitions (e.g., different channels, different Z
    # positions): the "acquisition time" of the timepoint is the start of the first
    # acquisition.
    return np.min(filtered_acquisition_times)


def time_between_subblocks(
    czi: CziFile, current_scene: int, start_frame: int, end_frame: int
) -> Optional[float]:
    """
    Calculates the time from the first acquisition of start_frame to the first
    acquisition of end_frame in milliseconds. Only the given scene is considered.
    """
    start_time = _acquisition_time(czi, current_scene, start_frame)
    end_time = _acquisition_time(czi, current_scene, end_frame)
    if start_time is None or end_time is None:
        return None
    delta = end_time - start_time
    # Difference in milliseconds is delta divided by 1 millisecond
    # Explicit conversion to float is required, lest we get a numpy float64
    return float(delta / np.timedelta64(1, "ms"))



from bioio_czi import Reader
from bioio_czi.aicspylibczi_reader.reader import Reader as AicsPyLibCziRwReader
from bioio_czi.pylibczirw_reader.reader import Reader as PylibCziRwReader

from .conftest import LOCAL_RESOURCES_DIR


def test_use_aicspylibczi_true() -> None:
    # Arrange
    uri = LOCAL_RESOURCES_DIR / "S=2_4x2_T=2=Z=3_CH=2.czi"

    # Act
    reader = Reader(uri, use_aicspylibczi=True)

    # Assert
    assert isinstance(reader._implementation, AicsPyLibCziRwReader)


def test_use_aicspylibczi_false() -> None:
    # Arrange
    uri = LOCAL_RESOURCES_DIR / "S=2_4x2_T=2=Z=3_CH=2.czi"

    # Act: pylibczirw by default
    reader = Reader(uri)

    # Assert
    assert isinstance(reader._implementation, PylibCziRwReader)



import logging
from typing import Any, Dict, Optional
from xml.etree import ElementTree as ET

from bioio_base.dimensions import DimensionNames

from bioio_czi.bounding_box import size

from .metadata import generate_ome_channel_id

log = logging.getLogger(__name__)


def get_channel_names(
    xml: ET.Element, scene_index: int, dims_shape: Dict[str, Any]
) -> Optional[list[str]]:
    """
    Get the channel names for the given scene index.

    Parameters
    ----------
    metadata: xml.etree.ElementTree.Element
        The metadata to search for channel names.
    scene_index: int
    """
    # Get all images
    img_sets = xml.findall(".//Image/Dimensions/Channels")

    if len(img_sets) == 0:
        return None

    # Select the current scene
    img = img_sets[0]
    if scene_index < len(img_sets):
        img = img_sets[scene_index]

    # Construct channel name list
    scene_channel_list = []
    channels = img.findall("./Channel")
    number_of_channels_in_data = size(dims_shape, DimensionNames.Channel)

    # There may be more channels in the metadata than in the data
    # if so, we will just use the first N channels and log
    # a warning to the user
    if len(channels) > number_of_channels_in_data:
        log.warning(
            "More channels in metadata than in data "
            f"({len(channels)} vs. {number_of_channels_in_data})"
        )

    for i, channel in enumerate(channels[:number_of_channels_in_data]):
        # Id is required, Name is not.
        # But we prefer to use Name if it is present
        channel_name = channel.attrib.get("Name")
        channel_id = channel.attrib.get("Id")
        if channel_name is None:
            # Idea: we could try to find a channel name from
            # DisplaySetting/Channels/Channel
            channel_name = channel_id
        if channel_name is None:
            # This is actually an error because Id was required by the spec
            channel_name = generate_ome_channel_id(str(scene_index), str(i))

        scene_channel_list.append(channel_name)
    return scene_channel_list



import datetime
from typing import Any

import pytest

from bioio_czi import Reader

from .conftest import LOCAL_RESOURCES_DIR


# Test each each of the following files in both aicspylibczi and pylibczirw modes.
#   variable_per_scene_dims.czi
#   OverViewScan.czi
# S=2_4x2_T=2=Z=3_CH=2.czi is only tested in aicspylibczi mode since it is used mainly
# for testing duration and interval, which aren't available in pylibczirw mode.
@pytest.mark.parametrize(
    "use_aicspylibczi, filename, expected",
    [
        (
            True,
            "variable_per_scene_dims.czi",
            {
                "Binning": "1x1",
                "Column": "4",
                "Dimensions Present": "TCZYX",
                "Image Size C": 1,
                "Image Size T": 2,
                "Image Size X": 1848,
                "Image Size Y": 1248,
                "Image Size Z": 2,
                "Imaged By": "sara.carlson",
                "Imaging Datetime": datetime.datetime(
                    2020, 1, 18, 0, 16, 29, 771361, tzinfo=datetime.timezone.utc
                ),
                "Objective": "10x/0.45Air",
                "Pixel Size X": 0.5416666666666666,
                "Pixel Size Y": 0.5416666666666666,
                "Pixel Size Z": 2.23,
                "Position Index": 1,
                "Row": "4",
                "Timelapse": True,
                "Timelapse Interval": datetime.timedelta(milliseconds=59927.0),
                "Total Time Duration": datetime.timedelta(milliseconds=59927.0),
            },
        ),
        (
            True,
            "OverViewScan.czi",
            {
                "Binning": "Other",
                "Column": None,
                "Dimensions Present": "CMYX",
                "Image Size C": 1,
                "Image Size T": None,
                "Image Size X": 544,
                "Image Size Y": 440,
                "Image Size Z": None,
                "Imaged By": "M1SRH",
                "Imaging Datetime": datetime.datetime(
                    2016, 3, 11, 10, 23, 44, 925154, tzinfo=datetime.timezone.utc
                ),
                "Objective": "5x/0.35Air",
                "Pixel Size X": 4.5743626119409,
                "Pixel Size Y": 4.5743626119409,
                "Pixel Size Z": None,
                "Position Index": None,
                "Row": None,
                "Timelapse": False,
                "Timelapse Interval": None,
                "Total Time Duration": None,
            },
        ),
        (
            True,
            "S=2_4x2_T=2=Z=3_CH=2.czi",
            {
                "Binning": "1x1",
                "Column": None,
                "Dimensions Present": "HTCZMYX",
                "Image Size C": 2,
                "Image Size T": 2,
                "Image Size X": 256,
                "Image Size Y": 256,
                "Image Size Z": 3,
                "Imaged By": "M1SRH",
                "Imaging Datetime": datetime.datetime(
                    2021, 6, 15, 6, 14, 13, 823569, tzinfo=datetime.timezone.utc
                ),
                "Objective": "5x/0.35Air",
                "Pixel Size X": 0.4,
                "Pixel Size Y": 0.4,
                "Pixel Size Z": 1.0,
                "Position Index": None,
                "Row": None,
                "Timelapse": True,
                "Timelapse Interval": datetime.timedelta(milliseconds=19160.1933),
                "Total Time Duration": datetime.timedelta(milliseconds=19160.1933),
            },
        ),
        (
            False,
            "variable_per_scene_dims.czi",
            {
                "Binning": "1x1",
                "Column": "4",
                "Dimensions Present": "TCZYX",
                "Image Size C": 1,
                "Image Size T": 2,
                "Image Size X": 1848,
                "Image Size Y": 1248,
                "Image Size Z": 2,
                "Imaged By": "sara.carlson",
                "Imaging Datetime": datetime.datetime(
                    2020, 1, 18, 0, 16, 29, 771361, tzinfo=datetime.timezone.utc
                ),
                "Objective": "10x/0.45Air",
                "Pixel Size X": 0.5416666666666666,
                "Pixel Size Y": 0.5416666666666666,
                "Pixel Size Z": 2.23,
                "Position Index": 1,
                "Row": "4",
                "Timelapse": True,
                "Timelapse Interval": None,  # Available only in aicspylibczi mode
                "Total Time Duration": None,  # Available only in aicspylibczi mode
            },
        ),
        (
            False,
            "OverViewScan.czi",
            {
                "Binning": "Other",
                "Column": None,
                "Dimensions Present": "CYX",  # aicspylibczi mode has CMYX
                "Image Size C": 1,
                "Image Size T": None,
                # This image is larger in X and Y when using pylibczirw mode than
                # aicspylibczi mode because all the tiles are stitched together.
                "Image Size X": 7398,
                "Image Size Y": 3212,
                "Image Size Z": None,
                "Imaged By": "M1SRH",
                "Imaging Datetime": datetime.datetime(
                    2016, 3, 11, 10, 23, 44, 925154, tzinfo=datetime.timezone.utc
                ),
                "Objective": "5x/0.35Air",
                "Pixel Size X": 4.5743626119409,
                "Pixel Size Y": 4.5743626119409,
                "Pixel Size Z": None,
                "Position Index": None,
                "Row": None,
                "Timelapse": False,
                "Timelapse Interval": None,  # Available only in aicspylibczi mode
                "Total Time Duration": None,  # Available only in aicspylibczi mode
            },
        ),
    ],
)
def test_standard_metadata(
    use_aicspylibczi: bool, filename: str, expected: dict[str, Any]
) -> None:
    uri = LOCAL_RESOURCES_DIR / filename
    reader = Reader(uri, use_aicspylibczi=use_aicspylibczi)
    metadata = reader.standard_metadata.to_dict()

    # Compare each key's values.
    for key, expected_value in expected.items():
        error_message = f"{key}: Expected: {expected_value}, Actual: {metadata[key]}"
        if isinstance(expected_value, float):
            assert metadata[key] == pytest.approx(expected_value), error_message
        else:
            assert metadata[key] == expected_value, error_message


# These test cases are specifically to check that standard_metadata reports metadata
# of the user-selected scene.
@pytest.mark.parametrize(
    "use_aicspylibczi, filename, scene, expected",
    [
        (
            True,
            "variable_per_scene_dims.czi",
            0,
            {
                "Image Size T": 2,
                "Timelapse Interval": datetime.timedelta(milliseconds=59927.0),
                "Total Time Duration": datetime.timedelta(milliseconds=59927.0),
            },
        ),
        (
            True,
            "variable_per_scene_dims.czi",
            1,
            {
                "Image Size T": 1,
                "Timelapse Interval": None,
                "Total Time Duration": None,
            },
        ),
        (
            False,
            "variable_per_scene_dims.czi",
            0,
            {
                "Image Size T": 2,
            },
        ),
        (
            False,
            "variable_per_scene_dims.czi",
            1,
            {
                # This should be 1, but pylibczirw assumes all scenes have the same
                # shape, so this is a known defect of pylibczirw mode.
                "Image Size T": 2,
            },
        ),
    ],
)
def test_standard_metadata_with_set_scene(
    use_aicspylibczi: bool, filename: str, scene: int, expected: dict[str, Any]
) -> None:
    # Arrange
    uri = LOCAL_RESOURCES_DIR / filename
    reader = Reader(uri, use_aicspylibczi=use_aicspylibczi)

    # Act
    reader.set_scene(scene)
    metadata = reader.standard_metadata.to_dict()

    # Sanity check
    assert reader.current_scene_index == scene

    # Assert
    # Compare only values mentioned in "expected"
    for key, expected_value in expected.items():
        error_message = f"{key}: Expected: {expected_value}, Actual: {metadata[key]}"
        if isinstance(expected_value, float):
            assert metadata[key] == pytest.approx(expected_value), error_message
        else:
            assert metadata[key] == expected_value, error_message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path

import lxml.etree as ET
from ome_types import from_xml
import pytest

###############################################################################

TEST_DIR = Path(__file__).parent
RESOURCES_DIR = TEST_DIR / "resources"

REPO_DIR = TEST_DIR.parent


@pytest.fixture
def xslt_path():
    return str(REPO_DIR / "xslt" / "czi-to-ome.xsl")


###############################################################################


@pytest.mark.parametrize(
    "czi_xml_filename",
    [
        "s_3_t_1_c_3_z_5.czi.xml",
        "s_1_t_1_c_5_z_20.czi.xml",
        "s_3_t_1_c_3_z_5_with_subblocks.czi.xml",
        "s_1_t_1_c_5_z_20_with_subblocks.czi.xml",
        "subblocks.czi.xml",
        "RBG-8bit.czi.xml",
        "031816_Paxillin_2.czi.xml"
    ],
)
def test_transform(xslt_path: str, czi_xml_filename: str):
    # Get full path for CZI XML
    czi_xml_filepath = str(RESOURCES_DIR / czi_xml_filename)

    # Parse template and generate transform function
    template = ET.parse(xslt_path)
    transformer = ET.XSLT(template)

    # Parse CZI XML
    czixml = ET.parse(czi_xml_filepath)

    # Attempt to run transform
    try:
        omexml = transformer(czixml)

        # Uncomment this block to look at produced OME XML when manual testing
        # with open("produced.ome.xml", "w") as f:
        #     f.write(ET.tostring(omexml, pretty_print=True, encoding="unicode"))

    # Catch any exception
    except Exception as e:
        print("Error during transform")
        print("-" * 80)
        print("Full Log:")
        for entry in transformer.error_log:
            print(
                (
                    f"{entry.filename}: {entry.line}, "
                    f"{entry.column}> {entry.message}>"
                )
            )
        raise e

    # Test read from OME-Types
    from_xml(str(omexml))


@pytest.mark.parametrize(
    "czi_xml_filename, expected_lasers",
    [
        (
            "Confocal_test_4c.czi.xml",
            [
                '<ome:Laser ID="LightSource:MTBLKM980LaserLine561"/>',
                '<ome:Laser ID="LightSource:MTBLKM980LaserLine488"/>',
                '<ome:Laser ID="LightSource:MTBLKM980LaserLine405"/>',
                '<ome:Laser ID="LightSource:MTBLKM980LaserLine639"/>',
            ],
        )
    ],
)
def test_empty_wavelength_transform(xslt_path: str, czi_xml_filename: str, expected_lasers: list):
    # Arrange
    czi_xml_filepath = str(RESOURCES_DIR / czi_xml_filename)
    template = ET.parse(xslt_path)
    transformer = ET.XSLT(template)
    czixml = ET.parse(czi_xml_filepath)

    # Act
    omexml = transformer(czixml)
    omexml_str = ET.tostring(omexml, pretty_print=True, encoding="unicode")

    # Assert
    for laser in expected_lasers:
        assert laser in omexml_str, f"Expected laser {laser} not found in transformed XML"

    # Test that the resulting XML can be read by OME-Types
    from_xml(omexml_str)


# TODO: The way test files are downloaded needs to be looked into. At the very least
# test files for this repository need to be given their own S3 bucket or folder

#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import sys
import traceback
from pathlib import Path

from quilt3 import Package

###############################################################################

logging.basicConfig(
    level=logging.INFO,
    format="[%(levelname)4s: %(module)s:%(lineno)4s %(asctime)s] %(message)s",
)
log = logging.getLogger(__name__)

###############################################################################
# Args


class Args(argparse.Namespace):
    def __init__(self):
        self.__parse()

    def __parse(self):
        # Setup parser
        p = argparse.ArgumentParser(
            prog="download_test_resources",
            description=(
                "Download files used for testing this project. This will download "
                "all the required test resources and place them in the "
                "`tests/resources` directory."
            ),
        )

        # Arguments
        p.add_argument(
            "--top-hash",
            # Generated package hash from upload_test_resources
            default=None,
            help=(
                "A specific version of the package to retrieve. "
                "If none, will read from the TEST_RESOURCES_HASH.txt file."
            ),
        )
        p.add_argument(
            "--debug",
            action="store_true",
            help="Show traceback if the script were to fail.",
        )

        # Parse
        p.parse_args(namespace=self) 


###############################################################################
# Build package


def download_test_resources(args: Args):
    # Try running the download pipeline
    try:
        # Get test resources dir
        resources_dir = (
            Path(__file__).parent.parent / "bioio_czi" / "tests" / "resources"
        ).resolve()
        resources_dir.mkdir(exist_ok=True)

        # Use or read top hash
        if args.top_hash is None:
            with open(Path(__file__).parent / "TEST_RESOURCES_HASH.txt", "r") as f:
                top_hash = f.readline().rstrip()
        else:
            top_hash = args.top_hash

        log.info(f"Downloading test resources using top hash: {top_hash}")

        # Get quilt package
        package = Package.browse(
            "bioio_czi/test_resources",
            "s3://bioio-dev-test-resources",
            top_hash=top_hash,
        )

        # Download
        package["resources"].fetch(resources_dir)

        log.info(f"Completed package download.")

    # Catch any exception
    except Exception as e:
        log.error("=============================================")
        if args.debug:
            log.error("\n\n" + traceback.format_exc())
            log.error("=============================================")
        log.error("\n\n" + str(e) + "\n")
        log.error("=============================================")
        sys.exit(1)


###############################################################################
# Allow caller to directly run this module (usually in development scenarios)

if __name__ == "__main__":
    args = Args()
    download_test_resources(args)


import logging
from typing import Optional
from xml.etree.ElementTree import Element

log = logging.getLogger(__name__)


def position_index(scene: str) -> Optional[int]:
    """
    Extracts the numeric position index from the current scene name.
    Returns
    -------
    Optional[int]
        The numeric part of the scene name.
        Returns None if parsing fails.
    """
    try:
        # Use only the first part before a "-" if present
        prefix = scene.split("-")[0]
        return int(prefix[1:])
    except (IndexError, ValueError) as exc:
        log.warning(
            "Failed to parse position index from scene name '%s': %s",
            scene,
            exc,
            exc_info=True,
        )
    except Exception as exc:
        log.warning("Unexpected error parsing position index: %s", exc, exc_info=True)

    return None


def _row_or_column(
    metadata: Element, current_scene_index: int, row_or_column: str
) -> Optional[str]:
    """
    Extracts the well row or index for the current scene.
    Returns
    -------
    Optional[str]
        The column index as a string. Returns None if not found.
    """
    try:
        scenes = metadata.findall(
            "Metadata/Information/Image/Dimensions/S/Scenes/Scene"
        )
        for scene in scenes:
            scene_index = scene.get("Index")
            if scene_index is not None and int(scene_index) == current_scene_index:
                shape = scene.find("Shape")
                if shape is not None:
                    index = shape.find(
                        "RowIndex" if row_or_column == "row" else "ColumnIndex"
                    )
                    if index is not None:
                        return index.text
    except Exception as exc:
        log.warning(
            f"Failed to extract well {row_or_column} index: %s", exc, exc_info=True
        )

    return None


def column(metadata: Element, current_scene_index: int) -> Optional[str]:
    """
    Extracts the well column index for the current scene.
    Returns
    -------
    Optional[str]
        The column index as a string. Returns None if not found.
    """
    return _row_or_column(metadata, current_scene_index, "column")


def row(metadata: Element, current_scene_index: int) -> Optional[str]:
    """
    Extracts the well row index for the current scene.
    Returns
    -------
    Optional[str]
        The row index as a string. Returns None if not found.
    """
    return _row_or_column(metadata, current_scene_index, "row")
